package widebase.rq.handler

import java.io. { ByteArrayInputStream, File, RandomAccessFile }
import java.sql.Timestamp

import msgpack.rpcx.handler.AuthHandler

import org.joda.time. {

  LocalDate,
  LocalDateTime,
  LocalTime,
  Minutes,
  Seconds,
  YearMonth

}

import scala.collection.mutable. { ArrayBuffer, Map }

import vario.data.Datatype
import vario.file.FileVariantWriter
import vario.filter.StreamFilter
import vario.filter.StreamFilter.StreamFilter
import vario.io.VariantWriter

import widebase.db.table.Table
import widebase.io.column.FileColumnSaver
import widebase.io.filter.MagicId
import widebase.rq.concurrent.Consumer
import widebase.rq.persistence.PersistenceWriter

/** Handles rollback tables.
 *
 * @author myst3r10n
 */
trait PersistenceHandler extends AuthHandler {

  val path: String
  val filter: StreamFilter
  protected var persistences = Map[String, PersistenceWriter]()
  protected var subscriptions = Map[String, ArrayBuffer[Consumer]]()

  /** Write records into rollback table.
   *
   * @param name of rollback table
   * @param bytes of records
   */
  protected def write(name: String, bytes: Array[Byte]) {

    val table = Table.fromBytes(bytes, filter)

    val partition = new LocalDate(table.columns.head.typeOf match {

      case Datatype.Long => table.columns.head.head.asInstanceOf[Long]
      case Datatype.Month => table.columns.head.head.asInstanceOf[YearMonth].toLocalDate(1).toDateMidnight.getMillis
      case Datatype.Date => table.columns.head.head.asInstanceOf[LocalDate].toDateMidnight.getMillis
      case Datatype.Time => table.columns.head.head.asInstanceOf[LocalTime].getMillisOfDay
      case Datatype.DateTime => table.columns.head.head.asInstanceOf[LocalDateTime].toDateTime.getMillis
      case Datatype.Timestamp => table.columns.head.head.asInstanceOf[Timestamp].getTime

    } ).toString("yyyy-MM-dd")

    if(persistences.contains(name)) {

      val persistence = persistences(name)

      if(partition != persistence.partition) {

        persistence.label.close
        persistence.columns.foreach(writer => writer.close)

        persistences -= name

        if(subscriptions.contains(name))
          subscriptions(name).foreach(consumer => consumer.endOfDay)

      }
    }

    if(!persistences.contains(name)) {

      // Write labels
      val saver = new FileColumnSaver(path)
      saver.save(name, ".d", table.labels, true)(partition)

      val persistence = new PersistenceWriter(partition)

      // Location
      val dir = path + "/" + partition + "/" + name

      // Init label writer
      val outChannel = new RandomAccessFile(dir + "/.d", "rw").getChannel
      outChannel.tryLock
      persistence.label = new FileVariantWriter(outChannel, StreamFilter.None)

      // Init column writers
      table.foreach { case (name, column) =>

        val file = new File(dir + "/" + name)

        if(file.exists)
          file.delete

        val channel = new RandomAccessFile(file.getPath, "rw").getChannel
        channel.tryLock
        persistence.columns +=
          new FileVariantWriter(channel, StreamFilter.None)

        val writer = persistence.columns.last

        // Write magic
        writer.mode = Datatype.String
        writer.write(MagicId.Column.toString, true)

        // Write column type
        writer.mode = Datatype.Byte
        writer.write(column.typeOf.id.toByte)

        // Write column length
        writer.mode = Datatype.Int
        writer.write(0)

        writer.mode = column.typeOf

      }

      persistences += name -> persistence

    }

    val persistence = persistences(name)
    persistence.records += 1

    var i = 0

    // Write column values
    table.columns.foreach { column => 

      column.head match {

        case value: Boolean => persistence.columns(i).write(value)
        case value: Byte => persistence.columns(i).write(value)
        case value: Char => persistence.columns(i).write(value)
        case value: Double => persistence.columns(i).write(value)
        case value: Float => persistence.columns(i).write(value)
        case value: Int => persistence.columns(i).write(value)
        case value: Long => persistence.columns(i).write(value)
        case value: Short => persistence.columns(i).write(value)
        case value: YearMonth => persistence.columns(i).write(value)
        case value: LocalDate => persistence.columns(i).write(value)
        case value: Minutes => persistence.columns(i).write(value)
        case value: Seconds => persistence.columns(i).write(value)
        case value: LocalTime => persistence.columns(i).write(value)
        case value: LocalDateTime => persistence.columns(i).write(value)
        case value: Timestamp => persistence.columns(i).write(value)
        case value: Symbol => persistence.columns(i).write(value)
        case value: String => persistence.columns(i).write(value)

      }

      i += 1

    }

    persistence.label.flush
    persistence.columns.foreach(writer => writer.flush)

    val offset = (MagicId.Column.toString.getBytes :+ (0: Byte)).size + 1

    persistence.columns.foreach { writer =>

      val backupMode = writer.mode

      writer.mode = Datatype.Byte
      writer.position = offset

      // Write column length
      writer.mode = Datatype.Int
      writer.write(persistence.records)

      writer.flush

      writer.mode = Datatype.Byte
      writer.position = writer.size

      writer.mode = backupMode

    }
  }
}

