package widebase.rq.handler

import java.io. { PrintWriter, StringWriter }

import msgpack.rpcx.handler.AuthHandler

import org.msgpack.rpc.Request

import scala.collection.mutable. { ArrayBuffer, Map }
import scala.concurrent.Lock

import widebase.rq.concurrent.Consumer

/** Handler to processing consumers.
 *
 * @author myst3r10n
 */
trait SubscribingHandler extends PersistenceHandler {

  /** Locks sensitive data. */
  protected val lock: Lock

  /** Map of consumers. */
  protected var consumers = Map[String, Consumer]()

  /** Publishing interval. */
  val interval: Int

  /** Receives next request from consumer.
   *
   * @param request of consumer
   * @param authId unique authentication id of consumer
   */
  def next(request: Request, authId: String) {

    try {

      isLoggedIn(authId)
      hasAuthorization(authId, "next")

      if(!consumers.contains(authId))
        throw SubscriptionNotFoundException(authId)

      consumers(authId).next(request)

    } catch {

      case e: Throwable =>
        val message = new StringWriter
        val printer = new PrintWriter(message)
        e.printStackTrace(printer)
        log.error(message.toString)
        request.sendError(e.toString)

    }
  }

  /** Receives rollback request from consumer.
   *
   * @param request of consumer
   * @param authId unique authentication id of consumer
   */
  def rollback(request: Request, authId: String) {

    try {

      isLoggedIn(authId)
      hasAuthorization(authId, "records")

      val map = new java.util.HashMap[String, String]

      map.put("records", consumers(authId).rollback.records.toString)
      map.put("partition", consumers(authId).rollback.partition)

      request.sendResult(map)

    } catch {

      case e: Throwable =>
        val message = new StringWriter
        val printer = new PrintWriter(message)
        e.printStackTrace(printer)
        log.error(message.toString)
        request.sendError(e.toString)

    }
  }

  /** Receives subscribe request from consumer.
   *
   * @param request of consumer
   * @param authId unique authentication id of consumer
   * @param name of table
   * @param selector to filter records
   */
  def subscribe(
    request: Request,
    authId: String,
    name: String,
    selector: String = null) {

    try {

      isLoggedIn(authId)
      hasAuthorization(authId, "subscribe")

      if(selector != null) {

        // selector code...

      }

      val consumer = new Consumer(authId, name, selector, interval)

      consumer.start

      if(consumers.contains(authId))
        consumers(authId) = consumer
      else
        consumers += authId -> consumer

      lock.acquire

      if(persistences.contains(name)) {

        consumer.rollback.records = persistences(name).records
        consumer.rollback.partition = persistences(name).partition
        

      } else {

        consumer.rollback.records = 0
        consumer.rollback.partition = ""

      }

      if(subscriptions.contains(name))
        subscriptions(name) += consumer
      else
        subscriptions += name -> ArrayBuffer(consumer)

      lock.release

      log.info("subscribe: " + name + " by " + loginMap(authId).username +
        " (" + authId + ")")

      request.sendResult(true)

    } catch {

      case e: Throwable =>
        val message = new StringWriter
        val printer = new PrintWriter(message)
        e.printStackTrace(printer)
        log.error(message.toString)
        request.sendError(e.toString)

    }
  }

  /** Receives unsubscribe request from consumer.
   *
   * @param request of consumer
   * @param authId unique authentication id of consumer
   */
  def unsubscribe(request: Request, authId: String) {

    try {

      isLoggedIn(authId)
      hasAuthorization(authId, "unsubscribe")

      if(consumers.contains(authId)) {

        val table = consumers(authId).table
        val subscribers = subscriptions(table)
        subscribers.remove(subscribers.indexWhere(consumer =>
          consumer.authId == authId))

        val consumer = consumers(authId)
        consumers -= authId
        consumer.shutdown

        log.info("unsubscribe: " + table + " by " + loginMap(authId).username +
          " (" + authId + ")")

      }

      request.sendResult(true)

    } catch {

      case e: Throwable =>
        val message = new StringWriter
        val printer = new PrintWriter(message)
        e.printStackTrace(printer)
        log.error(message.toString)
        request.sendError(e.toString)

    }

  }
}

